"""Defines a simple checkpoint manager.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import glob
import os
import time
from typing import Any, Callable, List, Optional, Union

from totypes import json_utils

SERIALIZE_FN = json_utils.json_from_pytree
DESERIALIZE_FN = json_utils.pytree_from_json


@dataclasses.dataclass
class CheckpointManager:
    """A simple checkpoint manager with an orbax-like API.

    Example usage is as follows:

        mngr = checkpoint.CheckpointManager(
            path="experiment/wid_0000",
            save_interval_steps=10,
            max_to_keep=1,
        )

        # Initialize from a checkpoint if one exists.
        if mngr.latest_step() is not None:
            latest_step = mngr.latest_step()
            (params, state, scalars) = mngr.restore(latest_step)
        else:
            latest_step = -1
            params = ...  # initial parameters
            state = optimizer.init(params)
            scalars = {}

        for i in range(latest_step + 1, steps):
            # Update parameters, state, and scalars.
            mngr.save((params, state, scalars))

        mngr.save((params, state, scalars), force_save=True)

    Attributes:
        path: The path where checkpoints are to be saved.
        save_interval_steps: The save interval, which defines the steps at which
            checkpoints are saved. At other steps, calls to `save` do nothing, unless
            `force_save` is `True`.
        max_to_keep: The maximum number of checkpoints to keep.
        serialize_fn: Function which serializes the pytree.
        deserialize_fn: Function which deserializes the pytree.
    """

    path: str
    save_interval_steps: int
    max_to_keep: int
    serialize_fn: Callable[[Any], str] = SERIALIZE_FN
    deserialize_fn: Callable[[str], Any] = DESERIALIZE_FN

    def __post_init__(self):
        """Validates of `CheckpointManager` attributes."""
        if not os.path.exists(self.path):
            raise ValueError(f"`path` does not exist, got {self.path}.")

    def latest_step(self) -> Optional[int]:
        """Return the latest checkpointed step, or `None` if no checkpoints exist."""
        return latest_step(self.path)

    def save(self, step: int, pytree: Any, force_save: bool = False) -> None:
        """Save a pytree checkpoint."""
        if (step + 1) % self.save_interval_steps != 0 and not force_save:
            return
        serialized = self.serialize_fn(pytree)
        temp_fname = f"{self.path}/temp_{str(int(time.time()))}.json"
        with open(temp_fname, "w") as f:
            f.write(serialized)
        os.rename(temp_fname, fname_for_step(self.path, step))
        steps = checkpoint_steps(self.path)
        steps.sort()
        steps_to_delete = steps[: -self.max_to_keep]
        for step in steps_to_delete:
            os.remove(fname_for_step(self.path, step))

    def restore(self, step: int) -> Any:
        """Restore a pytree checkpoint."""
        return load(self.path, step, deserialize_fn=self.deserialize_fn)


def latest_step(wid_path: str) -> Optional[int]:
    """Return the latest checkpointed step, or `None` if no checkpoints exist."""
    steps = checkpoint_steps(wid_path)
    steps.sort()
    return None if len(steps) == 0 else steps[-1]


def checkpoint_steps(wid_path: str) -> List[int]:
    """Return the chackpoint filename for the given step."""
    fnames = glob.glob(fname_for_step(wid_path, step="*"))
    return [step_for_fname(f) for f in fnames]


def load(
    wid_path: str,
    step: int,
    deserialize_fn: Callable[[str], Any] = DESERIALIZE_FN,
) -> Any:
    """Load the checkpoitn for the given step from the `wid_path`."""
    with open(fname_for_step(wid_path, step)) as f:
        data = f.read()
    return deserialize_fn(data)


def fname_for_step(wid_path: str, step: Union[int, str]) -> str:
    """Return the filename for the given step."""
    step_str = f"{step:04}" if isinstance(step, int) else str(step)
    return f"{wid_path}/checkpoint_{step_str}.json"


def step_for_fname(checkpoint_fname: str) -> int:
    """Return the step for the given checkpoint filename."""
    return int(checkpoint_fname.split("_")[-1][:-5])



"""Functions for loading and analyzing experiments.

Copyright (c) 2023 The INVRS-IO authors.
"""

import glob
import json
import os
from datetime import datetime
from typing import Any, Dict, List, Sequence, Tuple

import jax
import numpy as onp
import pandas as pd

from invrs_utils.experiment import checkpoint, experiment

WID = "wid"
REPLICA = "replica"
STEP = "step"
LATEST_STEP = "latest_step"
LATEST_TIME_UTC = "latest_time_utc"
TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

LOSS = "loss"
LOSS_MIN = "loss_min"
LOSS_MEAN = "loss_mean"
LOSS_P05 = "loss_p05"
LOSS_P10 = "loss_p10"
LOSS_P25 = "loss_p25"
LOSS_P50 = "loss_p50"

EVAL_METRIC = "eval_metric"
EVAL_METRIC_MAX = "eval_metric_max"
EVAL_METRIC_MEAN = "eval_metric_mean"
EVAL_METRIC_P95 = "eval_metric_p95"
EVAL_METRIC_P90 = "eval_metric_p90"
EVAL_METRIC_P75 = "eval_metric_p75"
EVAL_METRIC_P50 = "eval_metric_p50"

SUMMARY_INTERVAL = "summary_interval"

SCALARS = "scalars"
COMPLETED = "completed"

PREFIX_WID_COL = "wid."

PREFIX_WID_PATH = "wid_"
PREFIX_CHECKPOINT = "checkpoint_"


def summarize_experiment(
    experiment_path: str,
    summarize_intervals: Sequence[Tuple[int, int]],
) -> pd.DataFrame:
    """Generate a summary dataframe containing all work units in an experiment.

    See `summarize_work_unit` for details on the summary information.

    Args:
        experiment_path: The path containing saved experiment data.
        summarize_intervals: Sequence of `(lo, hi)` defining the intervals for which
            summary data is to be calculated.

    Returns:
        The dataframe containing the summary data.
    """
    dfs = []
    wid_paths = glob.glob(f"{experiment_path}/{PREFIX_WID_PATH}*")
    for wid_path in wid_paths:
        if checkpoint_exists(wid_path):
            wid_config, wid_df = load_work_unit_scalars(wid_path)
            for replica_id, wid_replica_df in wid_df.groupby(REPLICA):
                replica_summary_df = summarize_work_unit(
                    wid_config,
                    wid_replica_df,
                    summarize_intervals=summarize_intervals,
                )
                replica_summary_df[f"{PREFIX_WID_COL}{REPLICA}"] = replica_id
                dfs.append(replica_summary_df)
    return pd.concat(dfs)


def summarize_work_unit(
    wid_config: Dict,
    wid_df: pd.DataFrame,
    summarize_intervals: Sequence[Tuple[int, int]],
) -> pd.DataFrame:
    """Generates a dataframe summarizing an experiment work unit.

    For each summary interval, the following quantities are computed:
      - The minimum loss over the interval
      - The mean loss over the interval
      - The 5th, 10th, 25th, and 50th percentile of loss over the interal
      - The maximum eval metric over the interval (if `eval_metric` is a column in
        the work unit dataframe).
      - The mean eval metric over the interval
      - The 95th, 90th, 75th, and 50th percentile of eval metric over the interval

    Args:
        wid_config: Dictionary containing the work unit configuration. Config data
            is included in the columns of the output dataframe.
        wid_df: Dataframe containing the work unit scalars.
        summarize_intervals: Sequence of `(lo, hi)` defining the intervals for which
            summary data is to be calculated.

    Returns:
        The dataframe containing the summary data.
    """
    data: Dict[str, List[Any]] = {}

    loss = onp.asarray(wid_df[LOSS])
    for key in (
        SUMMARY_INTERVAL,
        LOSS_MIN,
        LOSS_MEAN,
        LOSS_P05,
        LOSS_P10,
        LOSS_P25,
        LOSS_P50,
    ):
        data[key] = []

    include_eval_metric = EVAL_METRIC in wid_df.columns
    if include_eval_metric:
        eval_metric = onp.asarray(wid_df[EVAL_METRIC])
        for key in (
            EVAL_METRIC_MAX,
            EVAL_METRIC_MEAN,
            EVAL_METRIC_P95,
            EVAL_METRIC_P90,
            EVAL_METRIC_P75,
            EVAL_METRIC_P50,
        ):
            data[key] = []

    for lo, hi in summarize_intervals:
        interval_loss = loss[lo : min(hi, len(loss))]
        data[SUMMARY_INTERVAL].append(f"{lo:03}-{hi:03}")

        data[LOSS_MIN].append(onp.amin(interval_loss))
        data[LOSS_MEAN].append(onp.mean(interval_loss))
        data[LOSS_P05].append(onp.percentile(interval_loss, 5))
        data[LOSS_P10].append(onp.percentile(interval_loss, 10))
        data[LOSS_P25].append(onp.percentile(interval_loss, 25))
        data[LOSS_P50].append(onp.percentile(interval_loss, 50))

        if include_eval_metric:
            interval_eval_metric = eval_metric[lo : min(hi, len(loss))]
            data[EVAL_METRIC_MAX].append(onp.amax(interval_eval_metric))
            data[EVAL_METRIC_MEAN].append(onp.mean(interval_eval_metric))
            data[EVAL_METRIC_P95].append(onp.percentile(interval_eval_metric, 95))
            data[EVAL_METRIC_P90].append(onp.percentile(interval_eval_metric, 90))
            data[EVAL_METRIC_P75].append(onp.percentile(interval_eval_metric, 75))
            data[EVAL_METRIC_P50].append(onp.percentile(interval_eval_metric, 50))

    df = pd.DataFrame.from_dict(data)
    for key, value in wid_config.items():
        df[f"{PREFIX_WID_COL}{key}"] = value
    return df


def checkpoint_exists(wid_path: str) -> bool:
    """Return `True` if a checkpoint file exists."""
    if not os.path.isfile(f"{wid_path}/{experiment.FNAME_WID_CONFIG}"):
        return False
    if not glob.glob(f"{wid_path}/{PREFIX_CHECKPOINT}*.json"):
        return False
    return True


def load_work_unit_scalars(wid_path: str) -> Tuple[Dict, pd.DataFrame]:
    """Loads the scalars for a work unit from an experiment.

    Args:
        wid_path: The path for the work unit.

    Returns:
        A dictionary containing the work unit configuration, and a dataframe
        containing the logged scalars.
    """
    assert checkpoint_exists(wid_path)
    latest_step: int = checkpoint.latest_step(wid_path)  # type: ignore[assignment]

    with open(f"{wid_path}/{experiment.FNAME_WID_CONFIG}") as f:
        wid_config = json.load(f)
    wid_config[WID] = wid_path.split("/")[-1]
    wid_config[COMPLETED] = os.path.isfile(f"{wid_path}/{experiment.FNAME_COMPLETED}")
    wid_config = flatten_nested(wid_config)
    wid_config[LATEST_STEP] = latest_step

    timestamp = os.path.getmtime(checkpoint.fname_for_step(wid_path, latest_step))
    wid_config[LATEST_TIME_UTC] = datetime.utcfromtimestamp(timestamp).strftime(
        TIME_FORMAT
    )

    latest_checkpoint = checkpoint.load(wid_path, latest_step)

    scalars = latest_checkpoint[SCALARS]
    scalars_shape = list(scalars.values())[0].shape
    num_steps = scalars_shape[0]
    num_replicas = 1 if len(scalars_shape) == 1 else scalars_shape[1]

    replica_scalars = {}
    for key, value in scalars.items():
        assert value.shape == scalars_shape
        replica_scalars[key] = value.flatten()

    replica_scalars[REPLICA] = onp.tile(
        onp.arange(num_replicas)[onp.newaxis, :], (num_steps, 1)
    ).flatten()

    replica_scalars[STEP] = onp.tile(
        onp.arange(num_steps)[:, onp.newaxis], (1, num_replicas)
    ).flatten()

    df = pd.DataFrame.from_dict(replica_scalars)
    return wid_config, df


def flatten_nested(nested_dict):
    """Flatten a nested dictionary."""
    path_with_leaves = jax.tree_util.tree_leaves_with_path(nested_dict)
    flat = {}
    for path, leaf in path_with_leaves:
        key = ".".join(
            [
                p.key if isinstance(p, jax.tree_util.DictKey) else str(p.idx)
                for p in path
            ]
        )
        flat[key] = leaf
    return flat



"""Functions related to a single optimization work unit.

Copyright (c) 2023 The INVRS-IO authors.
"""

import functools
import os
import time
import warnings
from typing import Any, Callable, Dict, Optional, Protocol, Tuple

import jax
import jax.numpy as jnp
from jax import tree_util

from invrs_utils.experiment import checkpoint

PyTree = Any


def run_work_unit(
    key: jax.Array,
    wid_path: str,
    challenge: "Challenge",
    optimizer: "Optimizer",
    steps: int,
    champion_requires_binary: bool = True,
    response_kwargs_fn: Callable[[int], Dict[str, Any]] = lambda _: {},
    save_interval_steps: int = 10,
    max_to_keep: int = 1,
    print_interval: Optional[int] = 300,
    num_replicas: int = 1,
) -> None:
    """Runs a work unit.

    The work unit uses the `optimizer` to optimize the `challenge`, completing the
    specified number of `steps` but potentially stopping early if target performance
    has been achieved.

    Checkpoints are saved to the `wid_path`. These include the optimizer state, scalars
    (such as loss, eval metric, and any scalar metrics), and quantities associated with
    the best solution so far encountered (the "champion"); the best solution is the one
    which has the highest binarization degree and highest eval metric.

    Args:
        key: Random number generator key used to initialize parameters.
        wid_path: The path where work unit checkpoints are to be stored.
        challenge: The challenge to be optimized.
        optimizer: The optimizer for the work unit.
        steps: The number of optimization steps.
        champion_requires_binary: Determines if a new champion must have a greater
            degree of binarization than the previous champion. If `True`, champion
            results tend to become more binary, even at the cost of performance (loss).
        response_kwargs_fn: Function which computes keyword arguments to be supplied to
            the `challenge.component.response` method, given the step number. This
            enables e.g. evaluation with random wavelengths at each step.
        save_interval_steps: The interval at which checkpoints are saved to `wid_path`.
        max_to_keep: The maximum number of checkpoints to keep.
        print_interval: Optional, the seconds elapsed between updates.
        num_replicas: The number of replicas for the work unit. Each replica is
            identical except for the random seed used to generate initial parameters.
    """
    if os.path.isfile(f"{wid_path}/completed.txt"):
        return
    if not os.path.exists(wid_path):
        raise ValueError(f"{wid_path} does not exist.")

    print(f"{wid_path} starting")

    # Create a basic checkpoint manager that can serialize custom types.
    mngr = checkpoint.CheckpointManager(
        path=wid_path, save_interval_steps=save_interval_steps, max_to_keep=max_to_keep
    )

    keys = jax.random.split(key, num_replicas)
    del key

    if mngr.latest_step() is not None:
        latest_step: int = mngr.latest_step()  # type: ignore[assignment]
        latest_checkpoint = mngr.restore(latest_step)
        state = latest_checkpoint["state"]
        scalars = latest_checkpoint["scalars"]
        champion_result = latest_checkpoint["champion_result"]
    else:
        latest_step = -1  # Next step is `0`.
        latent_params = jax.vmap(challenge.component.init)(keys)
        state = jax.vmap(optimizer.init)(latent_params)
        scalars = {}
        champion_result = {}

    @jax.jit
    @functools.partial(jax.vmap, in_axes=(None, 0))
    def _step_fn(step: int, state: Any) -> Any:
        def loss_fn(params: Any) -> Tuple[jnp.ndarray, Any]:
            response_kwargs = response_kwargs_fn(step)
            response, aux = challenge.component.response(params, **response_kwargs)
            loss = challenge.loss(response)
            eval_metric = challenge.eval_metric(response)
            metrics = challenge.metrics(response, params, aux)
            return loss, (response, eval_metric, metrics, aux)

        params = optimizer.params(state)
        (value, (response, eval_metric, metrics, aux)), grad = jax.value_and_grad(
            loss_fn, has_aux=True
        )(params)
        updated_state = optimizer.update(
            grad=grad, value=value, params=params, state=state
        )

        # Ensure the previous state has the proper tree structure. If the update is
        # corrupted due to presence of `nan` in the loss value or gradient, we may
        # return the previous state with no update.
        previous_state = tree_util.tree_unflatten(
            treedef=tree_util.tree_structure(updated_state),
            leaves=tree_util.tree_leaves(state),
        )

        grad_contains_nan = jnp.any(
            jnp.asarray(
                [jnp.any(jnp.isnan(leaf)) for leaf in tree_util.tree_leaves(grad)]
            )
        )
        skip_update = jnp.isnan(value) | grad_contains_nan
        state = jax.lax.cond(skip_update, lambda: previous_state, lambda: updated_state)
        return state, skip_update, (params, value, response, eval_metric, metrics, aux)

    last_print_time = time.time()
    last_print_step = latest_step

    def _log_scalar(name: str, value: jnp.ndarray) -> None:
        if name not in scalars:
            assert value.ndim == 1
            scalars[name] = jnp.zeros((0, value.size))
        scalars[name] = jnp.concatenate([scalars[name], value[jnp.newaxis, :]])

    for i in range(latest_step + 1, steps):
        t0 = time.time()
        (
            state,
            skip_update,
            (params, loss_value, response, eval_metric, metrics, aux),
        ) = _step_fn(i, state)
        t1 = time.time()

        for replica, skipped in enumerate(skip_update):
            if skipped:
                warnings.warn(
                    f"Skipped update for replica {replica} due to `nan` values at "
                    f"step {i}."
                )

        if print_interval is not None and (
            time.time() > last_print_time + print_interval
        ):
            print(
                f"{wid_path} is now at step {i} "
                f"({(t1 - last_print_time) / (i - last_print_step):.1f}s / step)"
            )
            last_print_time = time.time()
            last_print_step = i

        _log_scalar("loss", loss_value)
        _log_scalar("eval_metric", eval_metric)
        _log_scalar("step_time", jnp.full((num_replicas,), t1 - t0))
        for name, metric_value in metrics.items():
            if _is_scalar(metric_value, num_replicas):
                _log_scalar(name, metric_value)

        champion_result = _update_champion_result(
            candidate={
                "step": jnp.full((num_replicas,), i),
                "loss": loss_value,
                "eval_metric": eval_metric,
                "binarization_degree": metrics["binarization_degree"],
                "params": params,
                "response": response,
                "metrics": metrics,
                "aux": aux,
            },
            champion=champion_result,
            requires_binary=champion_requires_binary,
        )
        ckpt_dict = {
            "state": state,
            "scalars": scalars,
            "champion_result": champion_result,
        }
        mngr.save(i, ckpt_dict)

    mngr.save(i, ckpt_dict, force_save=True)
    with open(f"{wid_path}/completed.txt", "w"):
        os.utime(wid_path, None)

    print(f"{wid_path} finished")


def _update_champion_result(
    candidate: Dict[str, Any],
    champion: Dict[str, Any],
    requires_binary: bool,
) -> Dict[str, Any]:
    """Updates champion results."""
    if candidate["step"][0] == 0:
        return candidate

    assert candidate["loss"].ndim == 1
    num_replicas = candidate["loss"].size

    new_champ = []
    for i in range(num_replicas):
        # If the old champion has a `nan` eval metric.
        if jnp.isnan(champion["eval_metric"][i]):
            new_champ.append(True)
        # If binarization is not required/relevant, new champ if eval metric is higher.
        elif candidate["binarization_degree"] is None or not requires_binary:
            new_champ.append(candidate["eval_metric"][i] > champion["eval_metric"][i])
        # If binarization is required, new champ if binarization is greater.
        elif candidate["binarization_degree"][i] > champion["binarization_degree"][i]:
            new_champ.append(True)
        # If binarization is required, no new champ if binarization is lesser.
        elif candidate["binarization_degree"][i] < champion["binarization_degree"][i]:
            new_champ.append(False)
        # If binarization is exactly equal, new champ if loss is lower.
        else:
            new_champ.append(candidate["eval_metric"][i] > champion["eval_metric"][i])

    old_champion_leaves = tree_util.tree_leaves(champion)
    candidate_leaves = tree_util.tree_leaves(candidate)
    new_champion_leaves = [
        jnp.where(
            jnp.reshape(
                jnp.asarray(new_champ), (num_replicas,) + (1,) * (new.ndim - 1)
            ),
            new,
            old,
        )
        for new, old in zip(candidate_leaves, old_champion_leaves)
    ]

    return tree_util.tree_unflatten(
        tree_util.tree_structure(candidate), new_champion_leaves
    )


def _is_scalar(x: Any, num_replicas: int) -> bool:
    """Returns `True` if `x` is a scalar, i.e. a singleton for each replica."""
    return isinstance(x, jnp.ndarray) and x.shape == (num_replicas,)


# -----------------------------------------------------------------------------
# Protocols defining the `Challenge` and `Optimizer` objects.
# -----------------------------------------------------------------------------


class Component(Protocol):
    def init(self, key: jax.Array) -> PyTree:
        ...

    def response(self, params: PyTree) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        ...


class Challenge(Protocol):
    component: Component

    def loss(self, response: PyTree) -> jnp.ndarray:
        ...

    def eval_metric(self, response: PyTree) -> jnp.ndarray:
        ...

    def metrics(
        self, response: PyTree, Params: PyTree, aux: Dict[str, Any]
    ) -> Dict[str, Any]:
        ...


class InitFn(Protocol):
    def __call__(self, params: PyTree) -> PyTree:
        ...


class ParamsFn(Protocol):
    def __call__(self, state: PyTree) -> PyTree:
        ...


class UpdateFn(Protocol):
    def __call__(
        self, *, grad: PyTree, value: float, params: PyTree, state: PyTree
    ) -> PyTree:
        ...


class Optimizer(Protocol):
    init: InitFn
    update: UpdateFn
    params: ParamsFn



"""Functions that enable running an optimization experiment.

Copyright (c) 2023 The INVRS-IO authors.
"""

import json
import multiprocessing as mp
import os
import random
import time
import traceback
from typing import Any, Callable, Dict, Sequence, Tuple

FNAME_WID_CONFIG = "setup.json"
FNAME_COMPLETED = "completed.txt"


def run_experiment(
    experiment_path: str,
    sweeps: Sequence[Dict[str, Any]],
    work_unit_fn: Callable[[Any], Any],
    workers: int,
    dry_run: bool,
    randomize: bool,
) -> None:
    """Runs an experiment."""
    # Set up checkpointing directory.
    wid_paths = [f"{experiment_path}/wid_{i:04}" for i in range(len(sweeps))]

    # Print some information about the experiment.
    print(
        f"Experiment:\n"
        f"  worker count = {max(1, workers)}\n"
        f"  work unit count = {len(sweeps)}\n"
        f"  experiment path = {experiment_path}\n"
        f"Work units:"
    )
    for wid_path, kwargs in zip(wid_paths, sweeps):
        print(f"  {wid_path}: {kwargs}")

    path_and_kwargs = list(zip(wid_paths, sweeps))
    if randomize:
        random.shuffle(path_and_kwargs)

    if dry_run:
        return

    if workers == 1:
        _ = list(map(work_unit_fn, path_and_kwargs))
    else:
        with mp.Pool(processes=workers) as pool:
            _ = list(pool.imap_unordered(work_unit_fn, path_and_kwargs))


def work_unit_fn(fn: Any) -> Callable[[Tuple[str, Dict[str, Any]]], Any]:
    """Wraps `fn` for use with `run_experiment`.

    The first argument to `fn` should be named `wid_path` and be the path to which work
    unit data will be saved. Remaining arguments are work unit hyperparameters.

    The wrapped function takes a single tuple as input, with the first element being
    the work unit path, and the second element being a dictionary giving keyword
    arguments to `fn`.

    Args:
        fn: The function to be wrapped.

    Returns:
        The wrapped function.
    """

    def wrapped_fn(path_and_kwargs: Tuple[str, Dict[str, Any]]) -> None:
        """Wraps `run_work_unit` so that it can be called by `map`."""
        wid_path, kwargs = path_and_kwargs

        if os.path.isfile(f"{wid_path}/{FNAME_COMPLETED}"):
            return
        if not os.path.exists(wid_path):
            os.makedirs(wid_path)
        with open(f"{wid_path}/{FNAME_WID_CONFIG}", "w") as f:
            json.dump(kwargs, f, indent=4)

        try:
            return fn(wid_path=wid_path, **kwargs)
        except KeyboardInterrupt as e:
            raise e
        except Exception as e:
            print(f"Exception: {wid_path}")
            tb = "".join(traceback.format_tb(e.__traceback__))
            message = f"{e}\n{tb}"
            with open(f"{wid_path}/exception_{str(int(time.time()))}.log", "w") as f:
                f.write(message)

    return wrapped_fn



"""Functions that assist in defining experiments with multiple work units.

Copyright (c) 2023 The INVRS-IO authors.
"""

import itertools
from typing import Any, Dict, Sequence

Sweep = Sequence[Dict[str, Any]]

_zip = zip


def sweep(name: str, values: Sequence[Any]) -> Sweep:
    """Generate a list of dictionaries defining a sweep."""
    return [{name: v} for v in values]


def zip(*sweeps: Sweep) -> Sweep:
    """Zip sweeps of different variables."""
    return [_merge(*kw) for kw in _zip(*sweeps, strict=True)]


def product(*sweeps: Sweep) -> Sweep:
    """Return the Cartesian product of multiple sweeps."""
    return [_merge(*kw) for kw in itertools.product(*sweeps)]


def _merge(*vars: Dict[str, Any]) -> Dict[str, Any]:
    """Merge dictionaries defining sweeps of multiple variables."""
    d = {}
    for v in vars:
        d.update(v)
    return d
